/*-------------------------------------------------------------------------*\

   Context

   A context is basically a tokenizer with a variable pool and parsing
   methods.

   Copyright (C) 1996 by Michael A. Muller
      
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
      This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
      
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.      

\*-------------------------------------------------------------------------*/

#file Context

#intinclude Var
#intinclude PtrSList
#intinclude Tok

#impinclude iostream
#impinclude fstream
#impinclude strstream
#impinclude ctype
#impinclude String
#impinclude Var
#impinclude Value
#impinclude StrConst
#impinclude VarInst
#impinclude TableVar
#impinclude TableVal
#impinclude EqExpr
#impinclude SbstExpr
#impinclude NotExpr
#impinclude CatExpr
#impinclude FuncExpr
#impinclude AndExpr
#impinclude OrExpr
#impinclude StrVar
#impinclude string
#impinclude sys/stat
#impinclude unistd

class TableVar;
class Table;
class Var;
class Expr;
class StrVal;

#define LINE_NUM        getSourceName() << '(' << getLineNum() << "): "

class Context : public Tokenizer {

   public:

      enum { defaultCommandPrefix = '@', defaultVarPrefix = '$',
             defaultVarStart = '(', defaultVarEnd = ')' };

#if defined(__os2__) || defined(__OS2__)
      enum { pathSep = ';' };
#else
      enum { pathSep = ':' };
#endif

      class SpecialChars {
         public:
            char
               commandPrefix,
               varPrefix,
               varStart,
               varEnd;

            SpecialChars(char commandPrefix0 = defaultCommandPrefix,
                         char varPrefix0 = defaultVarPrefix,
                         char varStart0 = defaultVarStart,
                         char varEnd0 = defaultVarEnd
                         ) :
                  commandPrefix(commandPrefix0),
                  varPrefix(varPrefix0),
                  varStart(varStart0),
                  varEnd(varEnd0)
               {
               }
         }
   
   private:
      ostream &dest;
      Context *enclose;
      SpecialChars specialChars;
      
      PtrSList<Var> vars;

      static SpecialChars defaultSpecialChars;

      StrVal *locateIncludeFile(const char *fileName)
         {
         StrVal fullName("");
         ObjPtr<StrVal> includePath;
         
         // get the "INCLUDE" variable
         const Var *includeVar = lookUp("INCLUDE");
         if (!includeVar || includeVar->classId() != ClassIds::StrVar)
            includePath = new StrVal(":");
         else
            includePath = (StrVal*)((StrVar*)includeVar)->evaluate();

         int pathLen = includePath->length();
         
         // iterate over the contained paths
         int cur = 0, end = 0;
         while (cur < pathLen)
            {
            end = includePath->index(pathSep, cur);
            if (end == -1)
               end = pathLen;
            const SubString &path = includePath->at(cur, end - cur);
            String fullName;
            if (path.length())
               cat(path, "/", fileName, fullName);
            else
               fullName = fileName;
            
            // check to see if the file exists
            struct stat statInfo;
            if (!stat(fullName, &statInfo) && S_ISREG(statInfo.st_mode))
               return new StrVal(fullName);
            
            // search again from the next location
            cur = end + 1;
            }
         
         return 0;         
         }

   public:
   
      Context(Context *enclose0,
              istream &source,
              ostream &dest0,
              const char *sourceName,
              int lineNum,
              const SpecialChars &specialChars0 = defaultSpecialChars
              ) :
            Tokenizer(source, sourceName, lineNum),
            enclose(enclose0),
            dest(dest0),
            specialChars(specialChars0)
         {
         }

      Context(Context *enclose0, istream &source, const char *sourceName, 
              int lineNum = 1
              ) :
            Tokenizer(source, sourceName, lineNum),
            enclose(enclose0),
            dest(enclose0->dest),
            specialChars(enclose0->specialChars)
         {
         }

//      Context(Context *enclose0) :
//            Tokenizer(enclose0->source, enclose0->sourceName, 
//                      enclose0->getLineNum()),
//            enclose(enclose0),
//            dest(enclose0->dest),
//            specialChars(enclose0->specialChars)
//         {
//         }    

      Context(istream &source = cin,
              ostream &dest0 = cout,
              const char *sourceName = "<stdin>",
              int lineNum = 1,
              const SpecialChars &specialChars0 = defaultSpecialChars) :
            Tokenizer(source, sourceName, lineNum),
            dest(dest0),
            enclose(0),
            specialChars(specialChars0)
         {
         }

      const Var *lookUp(const char *varName) const
         {
         for (Pix p = vars.first(); p; vars.next(p))
            if (vars(p)->name() == varName)
               return vars(p);
               
         if (enclose)
            return enclose->lookUp(varName);
         else
            return 0;
         }

      // defines (or redefines) the orphan 'var' by adding it to its var list
      Context &define(Var *var)
         {
         for (Pix p = vars.first(); p; vars.next(p))
            if (vars(p)->name() == var->name())
               {
               vars.reset(p, var);
               return *this;
               }
         vars.append(var);
         
         return *this;
         }
      
      /*---------------------------------------------------------------------*\
         Redefines the orphan 'var'.  If 'var' is not in the current context
         it will recursively attempt to redefine it in all enclosing
         contexts.  This failing,  it will throw a non-fatal exception.
      \*---------------------------------------------------------------------*/
      Context &redefine(Var *var)
         {
         for (Pix p = vars.first(); p; vars.next(p))
            if (vars(p)->name() == var->name())
               {
               vars.reset(p, var);
               return *this;
               }
         
         if (enclose)
            return enclose->redefine(var);
         else
            {
            cerr << LINE_NUM << "Undefined variable in redef: \"" 
               << var->name() << '"' << endl;
            throw 0;
            }
         
         return *this;
         }
      
      // copies the receivers variable pool to 'other'
      Context &copyVarsTo(Context &other)
         {
         for (Pix p = vars.first(); p; vars.next(p))
            other.define(vars(p)->copy());
         return *this;
         }

      Token parseSomething()
         {
         Token tok = nextToken();
         if (tok.type == Token::end)
            {
            cerr << LINE_NUM << "End of file not expected" << endl;
            throw 1;
            }
         return tok;
         }

      Token parseIdentifier()
         {
         Token tok = parseSomething();
         if (tok.type != Token::identifier)
            {
            cerr << LINE_NUM << "Identifier expected, not " << tok << endl;
            throw 1;
            }
         return tok;
         }
         
      Token parseSymbol(char c = 0)
         {
         Token tok = parseSomething();
         if (tok.type != Token::symbol)
            {
            cerr << LINE_NUM << "Symbol expected, not " << tok << endl;
            throw 1;
            }
         else if (c && tok.symVal != c)
            {
            cerr << LINE_NUM << '\'' << c << '\'' << " expected, not " <<
               tok << endl;
            throw 1;
            }
         return tok;
         }
         
      Token parseString()
         {
         Token tok = parseSomething();
         if (tok.type != Token::string)
            {
            cerr << LINE_NUM << "String expected, not " << tok << endl;
            throw 1;
            }
         return tok;
         }

      void parseToNewline()
         {
         while (source && getChar() != '\n') ;
         }

      Expr *parseAtomicExpression()
         {
         Token tok = parseSomething();
         switch (tok.type)
            {
            case Token::string:
               return new StrConst(tok.stringVal);
            case Token::identifier:
               {
               const Var *var = lookUp(tok.identVal);
               if (!var)
                  {
                  cerr << LINE_NUM << "Undefined variable " << tok
                     << " found in expression.";
                  throw 1;
                  }
               return new VarInst(var);
               }
            case Token::symbol:
               if (tok.symVal == '!')
                  {
                  ObjPtr<Expr> expr(parseAtomicExpression());
                  try {
                     return new NotExpr(expr.release());
                     }
                  catch (int ex)
                     {
                     if (!ex)
                        {
                        cerr << LINE_NUM << "Not operator can only be used "
                                            "upon a string expression."
                           << endl;
                        throw 1;
                        }
                     else
                        throw;
                     }
                  return expr.release();
                  }
               else if (tok.symVal == '(')
                  {
                  ObjPtr<Expr> expr(parseExpression());
                  parseSymbol(')');
                  return expr.release();
                  }
                  
               // otherwise fall through to error ...
               
            default:
               cerr << LINE_NUM << tok << " was unexpected at this time."
                  << endl;
               throw 1;
               break;
            }
         }

      // returns an orphaned expression
      Expr *parseExpression()
         {
         ObjPtr<Expr> expr(parseAtomicExpression());
         while (1) {
            // get the next token (end of stream may be acceptable under some
            // circumstances).
            Token tok = nextToken();
            if (tok.type == Token::symbol)
               switch (tok.symVal) {
                  case '=':
                     {
                     ObjPtr<Expr> expr2(parseExpression());
                     try {
                        expr = new EqExpr(expr.release(), expr2.release());
                        }
                     catch (int ex)
                        {
                        if (!ex) {
                           cerr << LINE_NUM << "Equality operator can only be used "
                              "between string expressions." << endl;
                           throw 1;
                           }
                        else
                           throw;
                        }
                     break;
                     }
                     
                  case '/':
                     {
                     ObjPtr<Expr> expr2(parseExpression());
                     try {
                        expr = new SbstExpr(expr.release(), expr2.release()); 
                        }
                     catch (int ex)
                        {
                        if (!ex) {
                           cerr << LINE_NUM << "Substring operator can only be used "
                              "between string expressions." << endl;
                           throw 1;
                           }
                        else
                           throw;
                        }
                     break;
                     }

                  case '&':
                     {
                     ObjPtr<Expr> expr2(parseExpression());
                     try {
                        expr = new AndExpr(expr.release(), expr2.release()); 
                        }
                     catch (int ex)
                        {
                        if (!ex) {
                           cerr << LINE_NUM << "And operator can only be used "
                              "between string expressions." << endl;
                           throw 1;
                           }
                        else
                           throw;
                        }
                     break;
                     }
                  case '|':
                     {
                     ObjPtr<Expr> expr2(parseExpression());
                     try {
                        expr = new OrExpr(expr.release(), expr2.release()); 
                        }
                     catch (int ex)
                        {
                        if (!ex) {
                           cerr << LINE_NUM << "Or operator can only be used "
                              "between string expressions." << endl;
                           throw 1;
                           }
                        else
                           throw;
                        }
                     break;
                     }

                  case '.':
                     {
                     ObjPtr<Expr> expr2(parseExpression());
                     try {
                        expr = new CatExpr(expr.release(), expr2.release()); 
                        }
                     catch (int ex)
                        {
                        if (!ex) {
                           cerr << LINE_NUM << "Concatenation operator can only be used "
                              "between string expressions." << endl;
                           throw 1;
                           }
                        else
                           throw;
                        }
                     break;
                     }
                     
                  case '(':
                     {
                     PtrSList<Expr> parms;
                     while (1)
                        {
                        ObjPtr<Expr> expr2(parseExpression());
                        parms.append(expr2.release());
                        tok = nextToken();
                        if (tok.type == Token::symbol)
                           if (tok.symVal == ')')
                              break;
                           else if (tok.symVal == ',')
                              continue;
                        
                        cerr << LINE_NUM << "Functions must contain one or "
                           "more parameters seperated by commas and "
                           "terminated by a paren." << endl;
                        throw 1;
                        }
                        
                     try {
                        expr = new FuncExpr(this, expr.release(),
                                            parms); 
                        }
                     catch (int ex)
                        {
                        if (!ex) {
                           cerr << LINE_NUM << "Macro invocation operator () "
                              "can only be applied to a macro." << endl;
                           throw 1;
                           }
                        else
                           throw;
                        }
                     break;
                     }

// I've concluded that the only way to implement this is through a radical
// redesign of grunt.  The problem is that variables are bound to their
// expressions when parsed.  But the "to expand" value in an index
// expression must resolve its variables within the nested context of 
// a particular row.
#ifdef THE_WORLD_HAS_CHANGED
                  case '[':
                     {
                     ObjPtr<Expr> index(parseExpression());
                     ObjPtr<Expr> toExpand;
                     Token tok = parseSymbol();
                     if (tok.symVal == ';')
                        {
                        toExpand = parseExpression();
                        parseSymbol(']');
                        }
                     else if (tok.symVal != ']')
                        {
                        cerr << LINE_NUM << "']' or ';' expected, not " <<
                           tok << endl;
                        throw 1;
                        }
                     try {
                        expr = new IndxExpr(expr.release(),
                                            index.release(),
                                            toExpand.release()
                                            );
                        }
                     catch (int ex)
                        {
                        if (!ex) {
                           cerr << LINE_NUM << "Index operator "
                              "must be applied to a table with a string "
                              "index." << endl;
                           throw 1;
                           }
                        else
                           throw;
                        }
                     break;
                     }
#endif
                                    
                  case '{':
                  case ')':
                  case ';':
                  case ',':
                     putBack(tok.symVal);
                     return expr.release();
                     
                  default:
                     cerr << LINE_NUM << tok << " was unexpected at this time." << endl;
                     throw 1;
                  }
            else if (tok.type == Token::end)
               {
               return expr.release();
               }
            else
               {
               cerr << LINE_NUM << tok << " was unexpected at this time." << endl;
               throw 1;
               }
            }
         }

      // parses the input up until the matching '}', returns all text
      // between this point and that.  if 'terminateWithSemicolon' is true,
      // replaces the closing '}' with a semicolon; otherwise, discards the
      // terminating '}'
      String parseToMatch(bool terminateWithSemicolon = true)
         {
         ostrstream out;
         int state = 0;
         Token tok;
         char c;
         while (source && state >= 0) {
         
            // get a token for all non-zero (command) states, otherwise get
            // a character
            if (state)
               tok = parseSomething();
            else {
               c = getChar();
               }
               
            switch (state) {
               case 0:
                  if (c == specialChars.commandPrefix)
                     state = 1;
                  else if (c == specialChars.varPrefix)
                     // do special processing for embedded expressions,
                     // as we don't want to pick up commands in those.
                     state = 3;
                  out << c;
                  break;
                  
               // parse the token after the command prefix, if it's a '}',
               // we're done, if it's a # parse a comment and continue
               case 1:
                  if (tok.type == Token::symbol)
                     if (tok.symVal == '}')
                        {
                        state = -1;
                        if (terminateWithSemicolon)
                           out << ';';
                        break;
                        }
                     else if (tok.symVal == '#')
                        {
                        // discard the comment
                        parseToNewline();
                        
                        // make it a null statement
                        out << ';';
                        
                        // go back into pass-thru mode
                        state = 0;
                        break;
                        }
                     
                  // if not the '}' or '#', fall through to statement
                  // processing
               
               // parse some statement, just echo the tokens to 'out'
               case 2:
                  out << ' ' << tok.asString();
                  if (tok.type == Token::symbol)
                     {
                     if (tok.symVal == '{')
                        {
                        out << parseToMatch(false) << '}';
                        state = 0;
                        }
                        
                     else if (tok.symVal == ';')
                        state = 0;
                     }
                     
                  break;
               
               // this state exists just so that we don't do anything
               // special for escaped command and var prefixes.
               case 3:
                  out << ' ' << tok.asString();
                  if (tok.type == Token::symbol &&
                      tok.symVal == specialChars.varStart
                      )
                     state = 4;
                  else
                     state = 0;
                  break;

               // states greater than 3 mean we're parsing an expression
               default:
                  out << ' ' << tok.asString();
                  if (tok.type == Token::symbol)
                     {
                     // handle parenthesized sub-expressions.
                     if (tok.symVal == '(')
                        state++;
                     else if (state > 4 && tok.symVal == ')')
                        state--;
                     else if (state == 4 && tok.symVal == specialChars.varEnd)
                        state = 0;
                     }
                  break;
               
               }
            }
         
         out << '\0';
         
         char *buffer = out.str();
         String rval(buffer);
         delete [] buffer;
         return rval;
         }            

      void parseAllStatement()
         {
         // get the table in question
         Expr *expr = parseExpression();
         
         // begin parsing block
         parseSymbol('{');
         
         // stash the line number (we'll need it when parsing the body)
         int startLineNumber = lineNum;
         
         // and get the body of the statement
         String body = parseToMatch();
         
         Value *val = expr->evaluate();
         
         // make sure that value is a table
         if (val->classId() != ClassIds::TableVal)
            {
            cerr << LINE_NUM << "expression in 'all' statement must evaluate "
                << "to a table.";
            throw 1;
            }

         TableVal *tval = (TableVal*)val;
         int index = 0;
         for (Pix p = tval->first(); p; tval->next(p), index++)
            {
            istrstream source(body);
            Context nested(this, source, dest, sourceName, startLineNumber);
            Pix colHdr, col;
            const Row
               &headers = tval->headers(),
               &cur = (*tval)(p);
            
            // make all of the columns in this row a variable (named after
            // the column header) in the 'nested' context
            for (col = cur.first(), colHdr = headers.first();
                 col;
                 cur.next(col), headers.next(colHdr))
               {
               nested.define(new StrVar(headers(colHdr), cur(col)));
               }
           
            // define the special variable '_first' to be 'true' or ''
            // depending on whether or not this is the first row in the table.
            nested.define(new StrVar("_first", 
                                     (p == tval->first() ? "true" : "")
                                     )
                          );

            // define the special variable '_last' to be 'true' or ''
            // depending on whether or not this is the last row in the table.
            Pix q = p;
            tval->next(q);
            nested.define(new StrVar("_last",
                                     (q ? "" : "true")
                                     )
                          );

            // define the special variable '_index' to be the zero-based
            // index of the current row of the table.
            ostrstream out;
            out << index << '\0';
            char *indexStr = out.str();
            nested.define(new StrVar("_index", indexStr));
            delete [] indexStr;

            // evaluate the body
            nested.parseAll();
            }
         }

      void parseDefStatement()
         {
         Token tok = parseIdentifier();
         parseSymbol('=');
         ObjPtr<Expr> expr(parseExpression());
         parseSymbol(';');
         ObjPtr<Value> val(expr->evaluate());
         define(val->asVar(tok.identVal));
         }

      void parseRedefStatement()
         {
         Token tok = parseIdentifier();
         parseSymbol('=');
         ObjPtr<Expr> expr(parseExpression());
         parseSymbol(';');
         ObjPtr<Value> val(expr->evaluate());
         redefine(val->asVar(tok.identVal));
         }

      TableVar *parseTableHeaders(const char *name)
         {
         SLList<String> headers;
         while (1) {
            Token tok = parseSomething();
            switch (tok.type )
               {
               case Token::identifier:
                  headers.append(tok.identVal);
                  break;
               case Token::symbol:
                  if (tok.symVal != ';')
                     {
                     cerr << LINE_NUM << "Symbol " << tok
                        << " was unexpected at this time." << endl;
                     throw 2;
                     }
                  else
                     {
                     Row *headerRow = new Row(headers.length());
                     int i = 0;
                     for (Pix p = headers.first(); p; headers.next(p), i++)
                        (*headerRow)[i] = headers(p);
                     TableVar *var = new TableVar(name, headers.length());
                     var->headers(headerRow);
                     return var;
                     }
               default:
                  cerr << LINE_NUM << tok << " was unexpected at this time." << endl;
                  throw 2;
               }
            }
         }

      bool parseTableBody(Table *tab)
         {
         Row *row = 0;
         Pix col = 0;
         while (1) {
            Token tok = parseSomething();
            switch (tok.type) {
               case Token::symbol:
                  if (tok.symVal == '@')
                     {
                     parseSymbol('}');
                     if (col)
                        cerr << LINE_NUM << "Finished parsing table mid-row."
                           << endl;
                     return false;
                     }
                  else
                     {
                     cerr << LINE_NUM << tok << " was unexpected at this time."
                        << endl;
                     throw 2;
                     }
                  break;
               case Token::identifier:
               case Token::string:
               case Token::integer:
               case Token::floatingPoint:
                  if (!row)
                     {
                     row = &tab->newRow();
                     col = row->first();
                     }
                  switch (tok.type) {
                     case Token::identifier:
                        (*row)[col] = tok.identVal;
                        break;
                     case Token::string:
                        (*row)[col] = tok.stringVal;
                        break;
                     case Token::integer:
                        {
                        ostrstream out;
                        out << tok.intVal << '\0';
                        char *str = out.str();
                        (*row)[col] = str;
                        delete str;
                        break;
                        }
                     case Token::floatingPoint:
                        {
                        ostrstream out;
                        out << tok.floatVal << '\0';
                        char *str = out.str();
                        (*row)[col] = str;
                        delete str;
                        break;
                        }
                     }
                  row->next(col);
                  if (!col)
                     row = 0;
                  break;
               default:
                  cerr << LINE_NUM << tok << " was unexpected at this time."
                     << endl;
                  throw 2;
               }
            }
         }

      void parseTableStatement()
         {
         Token tableName = parseIdentifier();
         parseSymbol('{');
         TableVar *var = parseTableHeaders(tableName.identVal);           
         parseTableBody(var);
         define(var);
         }

      void parseIfStatement()
         {
         parseSymbol('(');
         ObjPtr<Expr> expr(parseExpression());
         parseSymbol(')');
         parseSymbol('{');
         int startingLineNum = lineNum;
         String body = parseToMatch();
         if (!expr->isStringExpr())
            {
            cerr << LINE_NUM << "Expression in \"if\" statement must evaluate "
               "to a string." << endl;
            throw 1;
            }
         ObjPtr<StrVal> val((StrVal*)expr->evaluate());
         if (val->length())
            {
            istrstream source(body);
            Context nested(this, source, dest, sourceName, startingLineNum);
            nested.parseAll();
            }
         }

      void parseIncludeStatement()
         {
         ObjPtr<Expr> fileNameExpr(parseExpression());
         if (!fileNameExpr->isStringExpr())
            {
            cerr << LINE_NUM << "Expression in \"include\" statement must "
               "evaluate to a string." << endl;
            throw 1;
            }
         ObjPtr<StrVal> fileName((StrVal*)fileNameExpr->evaluate());
         parseSymbol(';');
         
         // search the include path for the file
         ObjPtr<StrVal> fullName(locateIncludeFile(*fileName));
         if (!fullName)
            {
            cerr << LINE_NUM << "file: " << *fileName << " not found." << endl;
            throw 1;
            }
         
         // try to parse the file
         ifstream source(*fullName);
         if (source) {
            Context includedFile(this, source, dest, *fullName, 1);
            includedFile.parseAll();
            includedFile.copyVarsTo(*this);
            }
         else {
            cerr << LINE_NUM << "unable to read file: " << *fullName << endl;
            throw 1;
            }
         }

      void parseOutStatement()
         {
         ObjPtr<Expr> expr(parseExpression());
         parseSymbol('{');
         int startingLineNum = lineNum;
         String body = parseToMatch();
         if (!expr->isStringExpr())
            {
            cerr << LINE_NUM << "Expression in \"if\" statement must evaluate "
               "to a string." << endl;
            throw 1;
            }
         ObjPtr<StrVal> val((StrVal*)expr->evaluate());
         ofstream newDest(*val);
         if (newDest) {
            istrstream source(body);
            Context nested(this, source, newDest, sourceName, startingLineNum);
            nested.parseAll();
            }
         else {
            cerr << LINE_NUM << "unable to open \"" << *val << "\" for output."
               << endl;
            throw 1;
            }
         }

      // returns true on end statement
      bool parseStatement()
         {
         Token tok = parseSomething();
         switch (tok.type) {
            case Token::identifier:
               if (!strcmp(tok.identVal, "all"))
                  parseAllStatement();
               else if (!strcmp(tok.identVal, "end"))
                  return true;
               else if (!strcmp(tok.identVal, "def"))
                  parseDefStatement();
               else if (!strcmp(tok.identVal, "redef"))
                  parseRedefStatement();
               else if (!strcmp(tok.identVal, "table"))
                  parseTableStatement();
               else if (!strcmp(tok.identVal, "if"))
                  parseIfStatement();
               else if (!strcmp(tok.identVal, "include"))
                  parseIncludeStatement();
               else if (!strcmp(tok.identVal, "out"))
                  parseOutStatement();
               else
                  cerr << LINE_NUM << "Unrecocgnized keyword " << tok
                     << " found in input." << endl;
               break;
            case Token::symbol:
               if (tok.symVal == ';')
                  // null statement
                  return false;
               else if (tok.symVal == '#')
                  {
                  // comment
                  parseToNewline();
                  return false;
                  }
               else
                  {
                  cerr << LINE_NUM << "Symbol " << tok << " was not expected"
                     " at this time" << endl;
                  throw 1;
                  }
               break;
            default:
               cerr << LINE_NUM << tok << " was unexpected at this time.";
               throw 1;
            }
            
         return false;
         }

      void parseVarMain(Token tok)
         {
         const Var *var = lookUp(tok.identVal);
         if (var)
            {
            if (var->classId() == ClassIds::StrVar)
               dest << ((StrVar*)var)->val();
            else
               cerr << LINE_NUM << "Cannot expand non-scalar value."
                  << endl;
            }
         else
            cerr << LINE_NUM << "undefined variable "
               << tok << " referenced." << endl;
         }

      void parseVar()
         {
         Token tok;
         if (source) {
            tok = nextToken();
            if (tok.type == Token::symbol &&
                tok.symVal == specialChars.varStart)
               {
               ObjPtr<Expr> expr(parseExpression());
               if (expr->isStringExpr())
                  {
                  ObjPtr<StrVal> val((StrVal*)expr->evaluate());
                  dest << *val;
                  }
               else {
                  cerr << LINE_NUM << "Inlined expressions must evaluate to strings."
                     << endl;
                  }
                  
               tok = nextToken();
               if (tok.type != Token::symbol ||
                   tok.symVal != specialChars.varEnd)
                  cerr << LINE_NUM << '\'' 
                     << specialChars.varEnd << "' expected at end of var."
                     << endl;
               }
            else if (tok.type == Token::symbol &&
                        (tok.symVal == specialChars.varPrefix ||
                        tok.symVal == specialChars.commandPrefix)
                     )
               dest << tok.asString();
            else if (tok.type == Token::identifier)
               {
               parseVarMain(tok);
               }
            }
         else
            dest << specialChars.varPrefix;
         }
         
      void parseAll(bool needEndStatement = false)
         {
         source.unsetf(ios::skipws);
         while (source)
            {
            char c;
            c = getChar();
            if (!source) break;

            if (c == specialChars.commandPrefix)
               {
               if (parseStatement())
                  {
                  if (!needEndStatement)
                     cerr << LINE_NUM
                        << "'end' statement not needed for file!" << endl;
                  return;
                  }
               }
            else if (c == specialChars.varPrefix)
               parseVar();
            else
               dest << c;
            }
         if (needEndStatement)
            cerr << LINE_NUM << 
               "End of file found before expected 'end' statement!" << endl;
         }
   }


#>Context::SpecialChars Context::defaultSpecialChars;
