/*-------------------------------------------------------------------------*\

   grunt.cc

   the grunt main program.  Creates Context's for each file specfied on
   the command line and parses them, sending the results to standard output.

   Copyright (C) 1996 by Michael A. Muller
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
      This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
      
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.      

\*-------------------------------------------------------------------------*/

#ifdef __GNUG__
#include <typeinfo>
#endif

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include "Context.h"
#include "Var.h"
#include "StrVal.h"
#include "FuncVar.h"
#include "StrVar.h"

// ///////////////PRIMITIVE FUNCTIONS\\\\\\\\\\\\\\\\\\\\\\

#include "PFuncVal.h"
#include "strstream.h"

class Context;

Value *upper(Context *, const PtrSList<Value> &parms)
   {
   const Value *val = parms.front();
   if (val->classId() != ClassIds::StrVal)
      // XXX need a special exception for bad types in primitive functions XXX
      // XXX need to be able to indicate to the caller what went wrong     XXX
      // XXX May want to handle this by just writing errmsg to the context.XXX
      throw 1;
   
   ObjPtr<StrVal> rval(new StrVal(*(StrVal*)val));
   rval->upcase();
   return rval.release();
   }

Value *lower(Context *, const PtrSList<Value> &parms)
   {
   const Value *val = parms.front();
   if (val->classId() != ClassIds::StrVal)
      // XXX need a special exception for bad types in primitive functions XXX
      // XXX need to be able to indicate to the caller what went wrong     XXX
      // XXX May want to handle this by just writing errmsg to the context.XXX
      throw 1;
   
   ObjPtr<StrVal> rval(new StrVal(*(StrVal*)val));
   rval->downcase();
   return rval.release();
   }

Value *subst(Context *, const PtrSList<Value> &parms)
   {
   Pix p;
   int i;
   const Value *pparms[3];
   for (i = 0, p = parms.first(); p; ++i, parms.next(p))
      {
      pparms[i] = parms(p);
      if (pparms[i]->classId() != ClassIds::StrVal)
         // XXX need a special exception for bad types in primitive functions XXX
         // XXX need to be able to indicate to the caller what went wrong     XXX
         // XXX May want to handle this by just writing errmsg to the context.XXX
         throw 1;
      }

   ObjPtr<StrVal> rval(new StrVal(*(StrVal*)pparms[0]));
   rval->gsub(*(StrVal*)pparms[1], *(StrVal*)pparms[2]);
   return rval.release();
   }

Value *mygetenv(Context *, const PtrSList<Value> &parms)
   {
   const Value *val = parms.front();
   if (val->classId() != ClassIds::StrVal)
      // XXX need a special exception for bad types in primitive functions XXX
      // XXX need to be able to indicate to the caller what went wrong     XXX
      // XXX May want to handle this by just writing errmsg to the context.XXX
      throw 1;
   
   ObjPtr<StrVal> rval;
   const char *envVal = getenv(*(StrVal*)val);
   if (envVal)
      rval = new StrVal(envVal);
   else
      rval = new StrVal("");
   return rval.release();
   }

Value *eval(Context *outer, const PtrSList<Value> &parms)
   {
   const Value *val = parms.front();
   if (val->classId() != ClassIds::StrVal)
      // XXX need a special exception for bad types in primitive functions XXX
      // XXX need to be able to indicate to the caller what went wrong     XXX
      // XXX May want to handle this by just writing errmsg to the context.XXX
      throw 1;
   
   // create a new stringstream from the parameter so that we can parse it
   istrstream source(*(StrVal*)val);
   source.unsetf(ios::skipws);
   
   // create an output string stream to put the output into.
   ostrstream dest;
   
   // create a new context in which to do the parsing.  Use the current line
   // number because the real source (the first parameter) does not have a
   // line number.
   Context nested(outer, source, dest, outer->getSourceName(),
                  outer->getLineNum());
   
   // extract the first expression out of it. (anything following a
   // semicolon, end paren, or opening curly brace in the string is ignored).
   ObjPtr<Expr> expr(nested.parseExpression());
   
   return expr->evaluate();
   }

Value *expand(Context *outer, const PtrSList<Value> &parms)
   {
   const Value *val = parms.front();
   if (val->classId() != ClassIds::StrVal)
      // XXX need a special exception for bad types in primitive functions XXX
      // XXX need to be able to indicate to the caller what went wrong     XXX
      // XXX May want to handle this by just writing errmsg to the context.XXX
      throw 1;
   
   // create a new stringstream from the parameter so that we can parse it
   istrstream source(*(StrVal*)val);
   
   // create an output string stream to put the output into.
   ostrstream dest;
   
   // create a new context in which to do the parsing.  Use the current line
   // number because the real source (the first parameter) does not have a
   // line number.
   Context nested(outer, source, dest, outer->getSourceName(),
                  outer->getLineNum());

   // parse the expression
   nested.parseAll();
   
   // at this point "dest" should be full of the expanded output.  Turn that
   // into an StrVal
   dest << '\0';
   char *buf = dest.str();
   ObjPtr<StrVal> rval(new StrVal(buf));
   delete [] buf;
   
   return rval.release();
   }

// \\\\\\\\\\\\\\END OF PRIMITIVE FUNCTIONS////////////////

// command line option processors

void processCLInclude(Context &context, const char *path)
   {
   const Var *includeVar = context.lookUp("INCLUDE");
   ObjPtr<StrVal> includePath;
   if (includeVar && includeVar->classId() == ClassIds::StrVar)
      {
      includePath = (StrVal*)includeVar->evaluate();
      cat(*includePath, ":", path, *includePath);
      }
   else
      includePath = new StrVal(path);

      
   if (includeVar)
      context.redefine(new StrVar("INCLUDE", *includePath));
   else
      context.define(new StrVar("INCLUDE", *includePath));
   }

void processCLDefine(Context &context, const char *varDef)
   {
   String vdef = varDef, name = "", val = "";
   int endVar = vdef.index("=");
   if (endVar != -1)
      {
      name = vdef.at(0, endVar);
      val = vdef.after(endVar);
      }
   else
      {
      name = vdef;
      val = "true";
      }

   if (context.lookUp(name))
      context.redefine(new StrVar(name, val));
   else
      context.define(new StrVar(name, val));
   }

main(int argc, char **argv)
   {
   try {

      // define and populate the root context
      Context root(cin, cout, "<stdin>");
      root.define(new FuncVar("upper", PFuncVal("upper", upper, 1)));
      root.define(new FuncVar("lower", PFuncVal("lower", lower, 1)));
      root.define(new FuncVar("subst", PFuncVal("subst", subst, 3)));
      root.define(new FuncVar("getenv", 
                              PFuncVal("mygetenv", mygetenv, 1)
                              )
                  );
      root.define(new FuncVar("eval", PFuncVal("eval", eval, 1)));
      root.define(new FuncVar("expand", 
                              PFuncVal("expand", expand, 1)
                              )
                  );

      // flag to keep track of whether files were processed on the command 
      // line
      int processedFiles = 0;

      for (int i = 1; i < argc; i++)
         {    
         if (argv[i][0] == '-')
            {
            switch (argv[i][1])
               {
               case 'I':
                  if (strlen(argv[i]) > 2)
                     processCLInclude(root, &argv[i][2]);
                  else if (i + 1 == argc)
                     cerr << "The -I option requires an argument" << endl;
                  else
                     processCLInclude(root, argv[++i]);
                  break;
               case 'D':
                  if (strlen(argv[i]) > 2)
                     processCLDefine(root, &argv[i][2]);
                  else if (i + 1 == argc)
                     cerr << "The -I option requires an argument" << endl;
                  else
                     processCLDefine(root, argv[++i]);
                  break;
               default:
                  cerr << "Unknown option: " << argv[i] << endl;
               }
            }
         else
            {
            // argument is a file name
            ifstream in(argv[i]);
            if (in)
               {
               // parse the file into a child context
               Context child(&root, in, argv[i]);
               child.parseAll();
               child.copyVarsTo(root);
               processedFiles = 1;
               }
            else
               {
               cerr << "The file: \"" << argv[i] << "\" does not exist." 
                  << endl;
               }
            }
         }

      if (!processedFiles) {
         // parse standard input into a child context
         Context child(&root, cin, "<stdin>");
         child.parseAll();
         child.copyVarsTo(root);
         }
      }
   catch (int i)
      {
      cerr << "aborting: " << i << endl;
      exit(1);
      }
   catch (char *pc)
      {
      cout << "application threw a \"" << pc << '"' << endl;
      }
   catch ( ... )
      {
      cout << "uncaught exception thrown" << endl;
      }
   }
